/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  linux.c -- Nasty, non-portable, linux specific stuff which changes
             from kernel release to kernel release. ie the transparent 
             proxy calls.

***************************************/

#include <sys/utsname.h>
#include <sys/wait.h>
#include <signal.h>
#include "common.h"

#if HAVE_LINUX_NETFILTER_IPV4_H
# include <linux/netfilter_ipv4.h>
#endif

#if TRANS_DATA
#if USE_LIBIPTC
# include <libiptc.h>
# include <linux/netfilter_ipv4/ip_nat.h>
#endif
#endif

void get_kernel(void);

static enum {
	INIT,
	LINUX_2_0,
	LINUX_2_2,
	LINUX_2_4,
	OTHER
} kernel = INIT;


void get_kernel(void)
{
	struct utsname tmp;
	
	uname(&tmp);
	if(!strncmp(tmp.release,"2.0.",4)) kernel=LINUX_2_0;
	else if(!strncmp(tmp.release,"2.2.",4)) kernel=LINUX_2_2;
	else if(!strncmp(tmp.release,"2.4.",4)) kernel=LINUX_2_4;
	else kernel=OTHER;
}

/* ------------------------------------------------------------- **
**  Get the original destination address of a transparently proxied
**  socket.
**  ------------------------------------------------------------- */
int get_orig_dest(int fd, struct sockaddr_in *addr)
{
	socklen_t len;

	if(kernel==INIT) get_kernel();

	len=sizeof(*addr);
	switch(kernel){
	case LINUX_2_0:
	case LINUX_2_2:
		return( getsockname(fd, (struct sockaddr*) addr, &len) );
	default:
#ifdef SO_ORIGINAL_DST /*Header support for kernel 2.4*/
		return( getsockopt(fd, SOL_IP, SO_ORIGINAL_DST, 
				   (struct sockaddr*) addr, &len) );
#endif
		write_log(ERROR, "Running on a kernel we haven't been compiled for. Oooops.");
		return(-1);
	}
}

/* ------------------------------------------------------------- **
**  Get the address of the interface we connect to the client through
**  for putting in our 227 reply. For 2.4 do a getsockname on the
**  control socket. For 2.2 this gives us the orriginal destination of
**  the transparently proxied connection, so we do some nasty hackery
**  instead.
**  ------------------------------------------------------------- */
int get_local_address(struct sockaddr_in *addr)
{
	int sockfd;
        socklen_t len;

	/*If ListenAddress is in the config file then use the address
	 * from that*/
	*addr=config_getaddress();
	if(addr->sin_addr.s_addr!=0) {
		addr->sin_port=0;
		return(0);
	}

        if(kernel==INIT) get_kernel();
 
	switch(kernel){
	case LINUX_2_2:
		/* This piece of code ought to be taken out and shot
		**  (yes - it opens a UDP socket, does a getsockname,
		**  and then closes the socket before anything
		**  happens!) Suggestions for an alternative welcomed */
		
		if( (sockfd = socket(AF_INET, SOCK_DGRAM, 0))==-1) return(-1);

		addr->sin_family      = AF_INET;
		addr->sin_addr        = info->client_control.address.sin_addr;
		addr->sin_port        = htons(12345);
		len = sizeof(*addr);
			
		if(connect(sockfd, (struct sockaddr*) addr, len)==-1) {
			close(sockfd);
			return(-1);
		}
		
		if(getsockname(sockfd, (struct sockaddr*) addr, &len)==-1) {
			close(sockfd);
			return(-1);
		}

		close(sockfd);

		addr->sin_port=0;
		return(0);
	case LINUX_2_4:
	default:
                len=sizeof(*addr);
                return( getsockname(info->client_control.fd, (struct sockaddr*) addr, &len) );
	}
}

#if TRANS_DATA
/* ------------------------------------------------------------- **
**  Functions below are designed for the data link between the client
**  and the proxy. We want to fool the client that this connection
**  comes from the ftp server it connected to, so we have to be able
**  to either connect to the client with a false source address
**  (active mode), or intercept the client trying to connect to the
**  server's data port (passive mode).
**
**  On kernel 2.4 we do this using netfilter snat or dnat through
**  libiptc. This interface does not implement locking and so to
**  ensure only one of our processes intervenes at a time we fork a
**  process at startup to look after it. Also this process can run
**  with full privelges allowing the others to drop. On 2.2 we simply
**  do bind-to-foreign-address.[Not tested recently :) ]
**
**  Most of this stuff is a bit of a mess. Perhaps that is
**  unavoidable...
**  ------------------------------------------------------------- */

int sendrecv(mangle_req *req);
int init_chains(void);
static mangle_req saved;

/* ------------------------------------------------------------- **
**  Make a connection to dest which appears to have come from src. If
**  we can't then we just make the connection anyway.
**  ------------------------------------------------------------- */
int transp_connect(struct sockaddr_in dest, struct sockaddr_in src)
{
        int sockfd, i;
	mangle_req req;

        if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
                perror("socket");
                quit(SOCKET);
        }

	if(kernel==LINUX_2_2) {
		uid_t uid=geteuid();
		debug("Regaining priveliges for bind-to-foreign-address\n");
		seteuid(0);
		i=bind(sockfd, (struct sockaddr*)&src, sizeof(src));
		debug("Dropping them again\n");
		seteuid(uid);
	} else {
		struct sockaddr_in address;
		int lo,hi;
		lo=config_loport(ACTV);
		hi=config_hiport(ACTV);
		
		memset(&address, 0, sizeof(address));
		address.sin_family      = AF_INET;
		i=bind_me(sockfd, &address, lo, hi, hi-lo);
	}

	if(i) {
		debug("bind failed\n");
		close(sockfd);
		return(-1);
	}

	if(kernel==LINUX_2_4 && sufd>0){
		req.dest=dest;
		req.to=src;
		req.type=SNAT;
		req.pid=getpid();
		debug("Requesting SNAT...");
		sendrecv(&req);
		debug2("%s\n", req.type==SUCCEEDED ? "Success" : "Failed");
	}

	i = connect(sockfd, (struct sockaddr *)&dest, sizeof(dest));

	if(kernel==LINUX_2_4 && sufd>0){
		req.type=XSNAT;
		debug("Deleting SNAT...");
		sendrecv(&req);
		debug2("%s\n", req.type==SUCCEEDED ? "Success" : "Failed");
	}

        if (i) {
                close(sockfd);
                return(-1);
        }

        return(sockfd);

}

/* ------------------------------------------------------------- **
**  Listen in order to intercept any connections that are headed for
**  realdest, on a port from the range specified by use. You must call
**  il_free at some point before exiting to remove the iptables rule
**  this function adds under 2.4. Current implementation of il_free
**  means you can only have one intercept_listening socket per
**  process.
**  ------------------------------------------------------------- */
int intercept_listen(struct sockaddr_in intercept, struct sockaddr_in listen_on, socketuse use)
{
	int i, sockfd;

	il_free();
	sockfd = socket(AF_INET, SOCK_STREAM, 0);

	if(kernel==LINUX_2_2) {
		uid_t uid=geteuid();
		debug("Regaining privelidges for bind-to-foreign-address\n");
		seteuid(0);
		i=bind(sockfd, (struct sockaddr*)&intercept, sizeof(intercept));
		debug("Dropping them again\n");
		seteuid(uid);		
	} else {
		int lo,hi;

		if(sufd<1) return -1;
		lo=config_loport(use);
		hi=config_hiport(use);
		i=bind_me(sockfd, &listen_on, lo, hi, hi-lo);
	}

	if(i){
		debug("bind failed\n");
		close(sockfd);
		return(-1);
	}
	
	if(listen(sockfd, 5)){
		close(sockfd);
		return(-1);
	}

	if(use==CTRL){
		int i=1;
		setsockopt(sockfd, SOL_SOCKET, SO_OOBINLINE, &i, sizeof(i));
	}


	if(kernel==LINUX_2_4){
		mangle_req req;
		
		req.src.sin_addr.s_addr=INADDR_ANY;
		req.dest=intercept;
		req.to=listen_on;

		req.type=DNAT;
		req.pid=getpid();

		saved=req;
		saved.type=XDNAT;
		
		debug("Requesting DNAT...");
		if(sendrecv(&req)) {
			debug("Failed\n");
			return(-1);
		} else {
			debug("Success\n");
			return(sockfd);
		}
	} else return(sockfd);
}

/* ------------------------------------------------------------- **
**  Call this to remove the iptables rule introduced by a previous
**  intercept_listen.
**  ------------------------------------------------------------- */
int il_free(void)
{
	int i;

	if(kernel==LINUX_2_2 || sufd==-1) return(0);
	if(saved.pid==0) return(0);

	debug("Deleting DNAT...");
	i=sendrecv(&saved);
	debug2("%s\n", saved.type==SUCCEEDED ? "Success" : "Failed");

	saved.pid=0;
	
	return(i);
}

/* ------------------------------------------------------------- **
**  Send request, wait until we get a reply (addressed to us), and put
**  it in req.
**  ------------------------------------------------------------- */
int sendrecv(mangle_req *req)
{
#ifdef USE_LIBIPTC
	send(sufd, req, sizeof(*req), 0);

	do{
		recv(sufd, req, sizeof(*req), MSG_WAITALL | MSG_PEEK);
		if(req->pid!=getpid()) usleep(10);
	}while(req->pid!=getpid());
	recv(sufd, req, sizeof(*req), MSG_WAITALL);

	if(req->type==SUCCEEDED) return 0;
	else return -1;
#endif /*USE_LIBIPTC*/
}

/* ------------------------------------------------------------- **
**  Stuff below this is kernel 2.4 specific. One copy of this code
**  runs in a seperate process as root, and serves all iptables
**  requests. 
**  ------------------------------------------------------------- */

#ifdef USE_LIBIPTC

#define FROXSNAT "froxsnat"
#define FROXDNAT "froxdnat"

void serve_requests(int fd);
void req_fail(int fd, mangle_req *req);
static RETSIGTYPE suicide(int signo);
struct ipt_entry *get_entry(struct sockaddr_in src, struct sockaddr_in dst, 
			    struct sockaddr_in to, int snat);
#endif /*USE_LIBIPTC*/

/* ------------------------------------------------------------- **
**  Kernel 2.2: We return 0 to say everything is OK.
**  Kernel 2.4: 
**    If we aren't linked to libiptc return -1;
**    Otherwise Fork and return a socket in the child. We listen on
**    the other end in the parent (as root) and serve requests to add
**    rules to netfilter by all the other processes. We do the fork
**    this way so we can exit on SIGCHLD and not be left hanging
**    around after the main process has exited. **
**    ------------------------------------------------------------- */
int transdata_setup()
{
	int fds[2];
	
	get_kernel();
	if(kernel==LINUX_2_2) return(0);

#ifndef USE_LIBIPTC
	fprintf(stderr, 
"You appear to be running a 2.4.x Linux kernel, but frox was not configured\n"
"with --enable-libiptc. Data connections will NOT be transparently proxied\n");
	return(-1);
#else /*USE_LIBIPTC*/

	if(init_chains()==-1) {
		fprintf(stderr, 
"\nChains " FROXSNAT " and/or " FROXDNAT " do not exist. Data connections\n"
"will not be transparently proxied. Read README.transdata for details\n\n");
		return(-1);
	}

	saved.pid=0;
	if(socketpair(PF_UNIX, SOCK_DGRAM, 0, fds)==-1) return(-1);
	switch(fork()){
	case -1:
		close(fds[0]);
		close(fds[1]);
		return(-1);
	case 0:
		close(fds[0]);
		return(fds[1]);
	default:
		break;
	}
	close(fds[1]);
	signal(SIGCHLD, suicide);
	serve_requests(fds[0]);

	exit(0);
#endif /*USE_LIBIPTC*/
}

#ifdef USE_LIBIPTC
void serve_requests(int fd)
{
	struct ipt_entry *e;
	iptc_handle_t h;
	char chain[12];
	mangle_req req;

	do{
		recv(fd, &req, sizeof(req), MSG_WAITALL);
		
		switch(req.type){
		case SNAT:
		case XSNAT:
			req.src = config_getaddress();
			req.src.sin_port=0;
			/*FIXME make src.sin_port a permitted range*/
			strcpy(chain, FROXSNAT);
			break;
		case DNAT:
		case XDNAT:
			req.to.sin_addr = config_getaddress().sin_addr;
			/*FIXME check to.sin_port is reasonable*/
			strcpy(chain, FROXDNAT);
			break;
		case FLUSH:
			/* Flush our chains in case we missed
			 * something. Main accept loop sends us this
			 * whenever number of forks drops to 0. No
			 * reply needed.*/
			debug("Flushing our chains...");
			if((h = iptc_init("nat"))){
				iptc_flush_entries(FROXSNAT, &h);
				iptc_flush_entries(FROXDNAT, &h);
				if(iptc_commit(&h)) debug("Success\n");
				else debug("Failed\n");
			}
			continue;
		default:
			continue;
		}
		e = get_entry(req.src, req.dest, req.to, 
			      req.type==SNAT || req.type==XSNAT);
		if(e==NULL)  {
			req_fail(fd, &req);
			continue;
		}

		if( !(h = iptc_init("nat")) ) {
			free(e);
			req_fail(fd, &req);
			continue;
		}

		if(req.type==SNAT || req.type==DNAT){
			if(!iptc_append_entry(chain, e, &h))  {
				free(e);
				req_fail(fd, &req);
				continue;
			}
		} else {
			unsigned char *matchmask = malloc(e->next_offset);
			if(matchmask==NULL) {
				free(e);
				req_fail(fd, &req);
				continue;
			}
			
			if(!iptc_delete_entry(chain, e, matchmask, &h))  {
				free(e);
				free(matchmask);
				req_fail(fd, &req);
				continue;
			}
			free(matchmask);
		}
		if(!iptc_commit(&h)) {
			free(e);
			req_fail(fd, &req);
			continue;
		}
		
		free(e);
		req.type=SUCCEEDED;
		send(fd, &req, sizeof(req), 0);		
	}while(TRUE);
}

int init_chains()
{
	iptc_handle_t h;
	
	if( !(h = iptc_init("nat")) ) return(-1);
	if( !iptc_is_chain( FROXSNAT, h) ) return(-1);
	if( !iptc_is_chain( FROXDNAT, h) ) return(-1);
	return(0);
}

void req_fail(int fd, mangle_req *req)
{
	req->type=FAILED;
	send(fd, req, sizeof(*req), 0);
}

static RETSIGTYPE suicide(int signo)
{
	switch(signo){
	case SIGCHLD :
		waitpid(-1, (int *) 0, WNOHANG);
		exit(0);
	}
}

/* ------------------------------------------------------------- **
** Set up an ipt_entry structure. Which will do the equivalent of
** "iptables -p tcp -s src -d dst -j (SNAT|DNAT) --to to". If snat is
** TRUE we do snat, otherwise dnat. This probably isn't the correct
** way to use libiptc, but there isn't much sample code/documentation
** and I really couldn't face messing around with dlopen etc.
**
** The return value should be freed by the calling function.
**
** I want my bind-to-foreign-address back :)
** ------------------------------------------------------------- */
struct ipt_entry *get_entry(struct sockaddr_in src, struct sockaddr_in dst, 
			    struct sockaddr_in to, int snat)
{
	struct ipt_entry *e;

	struct ipt_entry_match *match;
	struct ipt_tcp *tcpinfo;

	struct ipt_entry_target *target;
        struct ip_nat_multi_range *mr;

	unsigned int size1, size2, size3;

	size1=IPT_ALIGN(sizeof(struct ipt_entry));
	size2=IPT_ALIGN(sizeof(struct ipt_entry_match) + 
			sizeof(struct ipt_tcp));
	size3=IPT_ALIGN(sizeof(struct ipt_entry_target) + 
			sizeof(struct ip_nat_multi_range));

	e = malloc(size1 + size2 + size3);
	if(e==NULL) {
		write_log(ERROR,"Malloc failure");
		return(NULL);
	}
	memset(e, 0, size1 + size2 + size3);

	/*Offsets to the other bits*/
	e->target_offset=size1 + size2;
	e->next_offset=size1 + size2 + size3;

	/*Set up packet matching rules*/
	if((e->ip.src.s_addr = src.sin_addr.s_addr) == INADDR_ANY) 
		e->ip.smsk.s_addr=0;
	else e->ip.smsk.s_addr = inet_addr("255.255.255.255");

	if((e->ip.dst.s_addr = dst.sin_addr.s_addr) == INADDR_ANY)
		e->ip.dmsk.s_addr=0;
	else e->ip.dmsk.s_addr = inet_addr("255.255.255.255");

	e->ip.proto = IPPROTO_TCP;
	e->nfcache = NFC_UNKNOWN; /*Think this stops caching.*/

	/*TCP specific matching(ie. ports)*/
	match=(struct ipt_entry_match*) e->elems;
	match->u.match_size=size2;
	strcpy(match->u.user.name,"tcp");	

	tcpinfo=(struct ipt_tcp*) match->data;

	if(src.sin_port==0){
		tcpinfo->spts[0]=ntohs(0);
		tcpinfo->spts[1]=ntohs(0xFFFF);
	} else tcpinfo->spts[0]=tcpinfo->spts[1]=ntohs(src.sin_port);
	if(dst.sin_port==0){
		tcpinfo->dpts[0]=ntohs(0);
		tcpinfo->dpts[1]=ntohs(0xFFFF);
	} else tcpinfo->dpts[0]=tcpinfo->dpts[1]=ntohs(dst.sin_port);

	/*And the target*/
	target=(struct ipt_entry_target*) (e->elems + size2);
	target->u.target_size=size3;
	if(snat) strcpy(target->u.user.name,"SNAT");
	else strcpy(target->u.user.name,"DNAT");

	mr=(struct ip_nat_multi_range*)target->data;
	mr->rangesize=1;

	mr->range[0].flags=IP_NAT_RANGE_PROTO_SPECIFIED |
		IP_NAT_RANGE_MAP_IPS;
	mr->range[0].min.tcp.port=mr->range[0].max.tcp.port=to.sin_port;
	mr->range[0].min_ip=mr->range[0].max_ip=to.sin_addr.s_addr;

	return e;
}
#endif /*USE_LIBIPTC*/
#endif /*TRANS_DATA*/
